@@ -24,11 +24,15 @@ class InvalidRouteNameException(Exception):
 
 
 class Client:
+    """
+    Client
+    """
     def __init__(
-        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
+        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, _id: str
     ):
         self.reader = reader
         self.writer = writer
+        self.id = _id
 
         self.requests: Dict[str, PendingRequest] = {}
 
@@ -124,7 +128,15 @@ class KARPServer:
         """
         pass
 
-    async def _handle(self, request, writer, _id, _ip):
+    def on_connection_lost(self, client: Client):
+        """
+        Called, when the connection gets lost
+        :param client:
+        :return:
+        """
+        pass
+
+    async def _handle(self, request, writer, _id, _ip) -> None:
         try:
             r = Utils.create_interaction_object(request)
             if isinstance(r, Request):
@@ -145,8 +157,15 @@ class KARPServer:
     ) -> None:
         _ip = writer.transport.get_extra_info("peername")
         _id = Request.generate_id()
-        self.clients[_id] = Client(reader, writer)
-        self.on_new_connection(self.clients[_id])
+        self.clients[_id] = Client(reader, writer, _id)
+        if asyncio.iscoroutinefunction(self.on_new_connection):
+            asyncio.ensure_future(self.on_new_connection(self.clients[_id]))
+        else:
+            asyncio.ensure_future(
+                asyncio.get_event_loop().run_in_executor(
+                    self.executor, self.on_new_connection, self.clients[_id]
+                )
+            )
         self.logger.info(f"[+] New Connection from {_ip}")
         while not self.stopped:
             try:
@@ -156,13 +175,25 @@ class KARPServer:
             if not requests:
                 break
 
-            for request in requests.decode().split("\n"):
-                if request:
-                    self.logger.debug(f"= {request} =")
-                    asyncio.ensure_future(
-                        self._handle(request, writer, _id, _ip)
-                    )
+            try:
+                for request in requests.decode().split("\n"):
+                    if request:
+                        self.logger.debug(f"= {request} =")
+                        asyncio.ensure_future(
+                            self._handle(request, writer, _id, _ip)
+                        )
+            except UnicodeDecodeError:
+                break
 
+        if asyncio.iscoroutinefunction(self.on_connection_lost):
+            asyncio.ensure_future(self.on_connection_lost(self.clients[_id]))
+        else:
+            asyncio.ensure_future(
+                asyncio.get_event_loop().run_in_executor(
+                    self.executor, self.on_connection_lost, self.clients[_id]
+                )
+            )
+        del self.clients[_id]
         self.logger.info(f"[-] Disconnected from {_ip}")
 
     def add_route(self, **kwargs) -> Callable:
@@ -191,4 +222,3 @@ class KARPServer:
         """
         self.stopped = True
         self.server.close()
-
